--- src/http/ngx_http_request.h.orig	Wed Jun 30 14:42:15 2010
+++ src/http/ngx_http_request.h	Wed Sep 22 08:25:06 2010
@@ -529,6 +529,10 @@
     u_char                           *header_name_end;
     u_char                           *header_start;
     u_char                           *header_end;
+    u_char                           *header_multiline_start;
+    u_char                           *header_multiline_end;
+
+    ngx_uint_t                        header_rewind;
 
     /*
      * a memory that can be reused after parsing a request line
--- src/http/ngx_http_parse.c.orig	Wed Jun 23 16:34:54 2010
+++ src/http/ngx_http_parse.c	Wed Sep 22 08:27:41 2010
@@ -749,7 +749,21 @@
     return NGX_OK;
 }
 
+void
+ngx_http_parse_header_multiline_save_current_line(ngx_http_request_t *r)
+{
+    if (r->header_multiline_start == NULL) {
+        r->header_multiline_start = r->header_start;
+        r->header_multiline_end = r->header_start;
+    } else {
+        *r->header_multiline_end++ = ' ';
+    }
 
+    r->header_multiline_end = ngx_copy(r->header_multiline_end,
+                                       r->header_start,
+                                       r->header_end - r->header_start);
+}
+
 ngx_int_t
 ngx_http_parse_header_line(ngx_http_request_t *r, ngx_buf_t *b,
     ngx_uint_t allow_underscores)
@@ -763,6 +777,7 @@
         sw_value,
         sw_space_after_value,
         sw_ignore_line,
+        sw_almost_done_or_multiline,
         sw_almost_done,
         sw_header_almost_done
     } state;
@@ -783,6 +798,11 @@
     hash = r->header_hash;
     i = r->lowcase_index;
 
+    if (r->header_rewind) {
+        r->header_rewind = 0;
+        b->pos--;
+    }
+
     for (p = b->pos; p < b->last; p++) {
         ch = *p;
 
@@ -791,6 +811,7 @@
         /* first char */
         case sw_start:
             r->header_name_start = p;
+            r->header_multiline_start = NULL;
             r->invalid_header = 0;
 
             switch (ch) {
@@ -883,6 +904,7 @@
         case sw_space_before_value:
             switch (ch) {
             case ' ':
+            case '\t':
                 break;
             case CR:
                 r->header_start = p;
@@ -909,11 +931,23 @@
                 break;
             case CR:
                 r->header_end = p;
-                state = sw_almost_done;
+                state = sw_almost_done_or_multiline;
                 break;
             case LF:
-                r->header_end = p;
-                goto done;
+                if (p + 1 >= b->last) {
+                    /* can't read ahead */
+                    r->header_rewind = 1;
+                } else if ((*(p + 1) == ' ') || (*(p + 1) == '\t')) {
+                    /* header is multi-line */
+                    r->header_end = p;
+                    ngx_http_parse_header_multiline_save_current_line(r);
+                    state = sw_space_before_value;
+                } else {
+                    /* header is single-line */
+                    r->header_end = p;
+                    goto done;
+                }
+                break;
             }
             break;
 
@@ -921,12 +955,24 @@
         case sw_space_after_value:
             switch (ch) {
             case ' ':
+            case '\t':
                 break;
             case CR:
-                state = sw_almost_done;
+                state = sw_almost_done_or_multiline;
                 break;
             case LF:
-                goto done;
+                if (p + 1 >= b->last) {
+                    /* can't read ahead */
+                    r->header_rewind = 1;
+                } else if ((*(p + 1) == ' ') || (*(p + 1) == '\t')) {
+                    /* header is multi-line */
+                    ngx_http_parse_header_multiline_save_current_line(r);
+                    state = sw_space_before_value;
+                } else {
+                    /* header is single-line */
+                    goto done;
+                }
+                break;
             default:
                 state = sw_value;
                 break;
@@ -944,6 +990,31 @@
             }
             break;
 
+        /* is it multi-line header? */
+        case sw_almost_done_or_multiline:
+            switch(ch) {
+            case LF:
+                if (p + 1 >= b->last) {
+                    /* can't read ahead */
+                    r->header_rewind = 1;
+                } else if ((*(p + 1) == ' ') || (*(p + 1) == '\t')) {
+                    /* header is multi-line */
+                    ngx_http_parse_header_multiline_save_current_line(r);
+                    state = sw_space_before_value;
+                } else {
+                    /* header is single-line, rewind */
+                    p--;
+                    state = sw_almost_done;
+                }
+                break;
+            default:
+                /* header is single-line, rewind */
+                p--;
+                state = sw_almost_done;
+                break;
+            }
+            break;
+
         /* end of header line */
         case sw_almost_done:
             switch (ch) {
@@ -980,6 +1051,17 @@
     r->state = sw_start;
     r->header_hash = hash;
     r->lowcase_index = i;
+
+    if (r->header_multiline_start) {
+        ngx_http_parse_header_multiline_save_current_line(r);
+
+        for (p = r->header_multiline_end; p < r->header_end; ) {
+            *p++ = '\0';
+        }
+
+        r->header_start = r->header_multiline_start;
+        r->header_end = r->header_multiline_end;
+    }
 
     return NGX_OK;
 
